@@ -211,18 +211,18 @@ from keystoneclient import exceptions
 from keystoneclient import session
 from oslo_config import cfg
 from oslo_serialization import jsonutils
-from oslo_utils import timeutils
 import six
+import webob.dec
 
 from keystonemiddleware.auth_token import _auth
 from keystonemiddleware.auth_token import _base
 from keystonemiddleware.auth_token import _cache
 from keystonemiddleware.auth_token import _exceptions as exc
 from keystonemiddleware.auth_token import _identity
+from keystonemiddleware.auth_token import _request
 from keystonemiddleware.auth_token import _revocations
 from keystonemiddleware.auth_token import _signing_dir
 from keystonemiddleware.auth_token import _user_plugin
-from keystonemiddleware.auth_token import _utils
 from keystonemiddleware.i18n import _, _LC, _LE, _LI, _LW
 
 
@@ -364,26 +364,6 @@ CONF.register_opts(_OPTS, group=_base.AUTHTOKEN_GROUP)
 
 _LOG = logging.getLogger(__name__)
 
-_HEADER_TEMPLATE = {
-    'X%s-Domain-Id': 'domain_id',
-    'X%s-Domain-Name': 'domain_name',
-    'X%s-Project-Id': 'project_id',
-    'X%s-Project-Name': 'project_name',
-    'X%s-Project-Domain-Id': 'project_domain_id',
-    'X%s-Project-Domain-Name': 'project_domain_name',
-    'X%s-User-Id': 'user_id',
-    'X%s-User-Name': 'username',
-    'X%s-User-Domain-Id': 'user_domain_id',
-    'X%s-User-Domain-Name': 'user_domain_name',
-}
-
-_DEPRECATED_HEADER_TEMPLATE = {
-    'X-User': 'username',
-    'X-Tenant-Id': 'project_id',
-    'X-Tenant-Name': 'project_name',
-    'X-Tenant': 'project_name',
-}
-
 
 class _BIND_MODE(object):
     DISABLED = 'disabled'
@@ -401,61 +381,6 @@ def _token_is_v3(token_info):
     return ('token' in token_info)
 
 
-def _get_token_expiration(data):
-    if not data:
-        raise exc.InvalidToken(_('Token authorization failed'))
-    if _token_is_v2(data):
-        return data['access']['token']['expires']
-    elif _token_is_v3(data):
-        return data['token']['expires_at']
-    else:
-        raise exc.InvalidToken(_('Token authorization failed'))
-
-
-def _confirm_token_not_expired(expires):
-    expires = timeutils.parse_isotime(expires)
-    expires = timeutils.normalize_time(expires)
-    utcnow = timeutils.utcnow()
-    if utcnow >= expires:
-        raise exc.InvalidToken(_('Token authorization failed'))
-
-
-def _v3_to_v2_catalog(catalog):
-    """Convert a catalog to v2 format.
-
-    X_SERVICE_CATALOG must be specified in v2 format. If you get a token
-    that is in v3 convert it.
-    """
-    v2_services = []
-    for v3_service in catalog:
-        # first copy over the entries we allow for the service
-        v2_service = {'type': v3_service['type']}
-        try:
-            v2_service['name'] = v3_service['name']
-        except KeyError:
-            pass
-
-        # now convert the endpoints. Because in v3 we specify region per
-        # URL not per group we have to collect all the entries of the same
-        # region together before adding it to the new service.
-        regions = {}
-        for v3_endpoint in v3_service.get('endpoints', []):
-            region_name = v3_endpoint.get('region')
-            try:
-                region = regions[region_name]
-            except KeyError:
-                region = {'region': region_name} if region_name else {}
-                regions[region_name] = region
-
-            interface_name = v3_endpoint['interface'].lower() + 'URL'
-            region[interface_name] = v3_endpoint['url']
-
-        v2_service['endpoints'] = list(regions.values())
-        v2_services.append(v2_service)
-
-    return v2_services
-
-
 def _conf_values_type_convert(conf):
     """Convert conf values into correct type."""
     if not conf:
@@ -507,6 +432,7 @@ class AuthProtocol(object):
         self._delay_auth_decision = self._conf_get('delay_auth_decision')
         self._include_service_catalog = self._conf_get(
             'include_service_catalog')
+        self._hash_algorithms = self._conf_get('hash_algorithms')
 
         self._identity_server = self._create_identity_server()
 
@@ -537,7 +463,6 @@ class AuthProtocol(object):
 
         self._check_revocations_for_cached = self._conf_get(
             'check_revocations_for_cached')
-        self._init_auth_headers()
 
     def _conf_get(self, name, group=_base.AUTHTOKEN_GROUP):
         # try config from paste-deploy first
@@ -546,37 +471,16 @@ class AuthProtocol(object):
         else:
             return CONF[group][name]
 
-    def _call_app(self, env, start_response):
-        # NOTE(jamielennox): We wrap the given start response so that if an
-        # application with a 'delay_auth_decision' setting fails, or otherwise
-        # raises Unauthorized that we include the Authentication URL headers.
-        def _fake_start_response(status, response_headers, exc_info=None):
-            if status.startswith('401'):
-                response_headers.extend(self._reject_auth_headers)
-
-            return start_response(status, response_headers, exc_info)
-
-        return self._app(env, _fake_start_response)
-
-    def __call__(self, env, start_response):
+    @webob.dec.wsgify(RequestClass=_request._AuthTokenRequest)
+    def __call__(self, request):
         """Handle incoming request.
 
         Authenticate send downstream on success. Reject request if
         we can't authenticate.
 
         """
-        def _fmt_msg(env):
-            msg = ('user: user_id %s, project_id %s, roles %s '
-                   'service: user_id %s, project_id %s, roles %s' % (
-                       env.get('HTTP_X_USER_ID'), env.get('HTTP_X_PROJECT_ID'),
-                       env.get('HTTP_X_ROLES'),
-                       env.get('HTTP_X_SERVICE_USER_ID'),
-                       env.get('HTTP_X_SERVICE_PROJECT_ID'),
-                       env.get('HTTP_X_SERVICE_ROLES')))
-            return msg
-
-        self._token_cache.initialize(env)
-        self._remove_auth_headers(env)
+        self._token_cache.initialize(request.environ)
+        request.remove_auth_headers()
 
         try:
             user_auth_ref = None
@@ -584,100 +488,59 @@ class AuthProtocol(object):
 
             try:
                 self._LOG.debug('Authenticating user token')
-                user_token = self._get_user_token_from_header(env)
-                user_token_info = self._validate_token(user_token, env)
-                user_auth_ref = access.AccessInfo.factory(
-                    body=user_token_info,
-                    auth_token=user_token)
-                env['keystone.token_info'] = user_token_info
-                user_headers = self._build_user_headers(user_auth_ref,
-                                                        user_token_info)
-                self._add_headers(env, user_headers)
+                user_token_info = self._get_user_token_from_request(request)
+                user_auth_ref, user_token_info = self._validate_token(
+                    user_token_info, request.environ)
+                request.environ['keystone.token_info'] = user_token_info
+                request.set_user_headers(user_auth_ref,
+                                         self._include_service_catalog)
             except exc.InvalidToken:
                 if self._delay_auth_decision:
                     self._LOG.info(
                         _LI('Invalid user token - deferring reject '
                             'downstream'))
-                    self._add_headers(env, {'X-Identity-Status': 'Invalid'})
+                    request.headers['X-Identity-Status'] = 'Invalid'
                 else:
                     self._LOG.info(
                         _LI('Invalid user token - rejecting request'))
-                    return self._reject_request(env, start_response)
+                    self._reject_request()
 
             try:
                 self._LOG.debug('Authenticating service token')
-                serv_token = self._get_service_token_from_header(env)
+                serv_token = request.headers.get('X-Service-Token')
                 if serv_token is not None:
-                    serv_token_info = self._validate_token(
-                        serv_token, env)
-                    serv_auth_ref = access.AccessInfo.factory(
-                        body=serv_token_info,
-                        auth_token=serv_token)
-                    serv_headers = self._build_service_headers(serv_token_info)
-                    self._add_headers(env, serv_headers)
+                    serv_auth_ref, serv_token_info = self._validate_token(
+                        serv_token, request.environ)
+                    request.set_service_headers(serv_auth_ref)
             except exc.InvalidToken:
                 if self._delay_auth_decision:
                     self._LOG.info(
                         _LI('Invalid service token - deferring reject '
                             'downstream'))
-                    self._add_headers(env,
-                                      {'X-Service-Identity-Status': 'Invalid'})
+                    request.headers['X-Service-Identity-Status'] = 'Invalid'
                 else:
                     self._LOG.info(
                         _LI('Invalid service token - rejecting request'))
-                    return self._reject_request(env, start_response)
+                    self._reject_request()
 
-            env['keystone.token_auth'] = _user_plugin.UserAuthPlugin(
-                user_auth_ref, serv_auth_ref)
+            p = _user_plugin.UserAuthPlugin(user_auth_ref, serv_auth_ref)
+            request.environ['keystone.token_auth'] = p
 
         except exc.ServiceError as e:
             self._LOG.critical(_LC('Unable to obtain admin token: %s'), e)
-            return self._do_503_error(env, start_response)
+            raise webob.exc.HTTPServiceUnavailable()
 
-        self._LOG.debug("Received request from %s", _fmt_msg(env))
+        if self._LOG.isEnabledFor(logging.DEBUG):
+            self._LOG.debug('Received request from %s' % p._log_format)
 
-        return self._call_app(env, start_response)
+        response = request.get_response(self._app)
 
-    def _do_503_error(self, env, start_response):
-        resp = _utils.MiniResp('Service unavailable', env)
-        start_response('503 Service Unavailable', resp.headers)
-        return resp.body
+        if response.status_int == 401:
+            response.headers.extend(self._reject_auth_headers)
 
-    def _init_auth_headers(self):
-        """Initialize auth header list.
+        return response
 
-        Both user and service token headers are generated.
-        """
-        auth_headers = ['X-Service-Catalog',
-                        'X-Identity-Status',
-                        'X-Service-Identity-Status',
-                        'X-Roles',
-                        'X-Service-Roles']
-        for key in six.iterkeys(_HEADER_TEMPLATE):
-            auth_headers.append(key % '')
-            # Service headers
-            auth_headers.append(key % '-Service')
-
-        # Deprecated headers
-        auth_headers.append('X-Role')
-        for key in six.iterkeys(_DEPRECATED_HEADER_TEMPLATE):
-            auth_headers.append(key)
-
-        self._auth_headers = auth_headers
-
-    def _remove_auth_headers(self, env):
-        """Remove headers so a user can't fake authentication.
-
-        Both user and service token headers are removed.
-
-        :param env: wsgi request environment
-
-        """
-        self._LOG.debug('Removing headers from request environment: %s',
-                        ','.join(self._auth_headers))
-        self._remove_headers(env, self._auth_headers)
-
-    def _get_user_token_from_header(self, env):
+    def _get_user_token_from_request(self, request):
         """Get token id from request.
 
         :param env: wsgi request environment
@@ -685,30 +548,21 @@ class AuthProtocol(object):
         :raises exc.InvalidToken: if no token is provided in request
 
         """
-        token = self._get_header(env, 'X-Auth-Token',
-                                 self._get_header(env, 'X-Storage-Token'))
+        token = request.headers.get('X-Auth-Token',
+                                    request.headers.get('X-Storage-Token'))
         if token:
             return token
         else:
             if not self._delay_auth_decision:
-                self._LOG.debug('Headers: %s', env)
+                self._LOG.debug('Headers: %s', dict(request.headers))
             raise exc.InvalidToken(_('Unable to find token in headers'))
 
-    def _get_service_token_from_header(self, env):
-        """Get service token id from request.
-
-        :param env: wsgi request environment
-        :returns: service token id or None if not present
-
-        """
-        return self._get_header(env, 'X-Service-Token')
-
     @property
     def _reject_auth_headers(self):
         header_val = 'Keystone uri=\'%s\'' % self._auth_uri
         return [('WWW-Authenticate', header_val)]
 
-    def _reject_request(self, env, start_response):
+    def _reject_request(self):
         """Redirect client to auth server.
 
         :param env: wsgi request environment
@@ -716,10 +570,43 @@ class AuthProtocol(object):
         :returns: HTTPUnauthorized http response
 
         """
-        resp = _utils.MiniResp('Authentication required',
-                               env, self._reject_auth_headers)
-        start_response('401 Unauthorized', resp.headers)
-        return resp.body
+        raise webob.exc.HTTPUnauthorized(body='Authentication required',
+                                         headers=self._reject_auth_headers)
+
+    def _token_hashes(self, token):
+        """Generate a list of hashes that the current token may be cached as.
+
+        With PKI tokens we have multiple hashing algorithms that we test with
+        revocations. This generates that whole list.
+
+        The first element of this list is the preferred algorithm and is what
+        new cache values should be saved as.
+
+        :param str token: The token being presented by a user.
+
+        :returns: list of str token hashes.
+        """
+        if cms.is_asn1_token(token) or cms.is_pkiz(token):
+            return list(cms.cms_hash_token(token, mode=algo)
+                        for algo in self._hash_algorithms)
+        else:
+            return [token]
+
+    def _cache_get_hashes(self, token_hashes):
+        """Check if the token is cached already.
+
+        Functions takes a list of hashes that might be in the cache and matches
+        the first one that is present. If nothing is found in the cache it
+        returns None.
+
+        :returns: token data if found else None.
+        """
+
+        for token in token_hashes:
+            cached = self._token_cache.get(token)
+
+            if cached:
+                return cached
 
     def _validate_token(self, token, env):
         """Authenticate user token
@@ -730,160 +617,80 @@ class AuthProtocol(object):
         :raises exc.InvalidToken: if token is rejected
 
         """
-        token_id = None
+        token_hashes = None
 
         try:
-            token_ids, cached = self._token_cache.get(token)
-            token_id = token_ids[0]
-            if cached:
-                # Token was retrieved from the cache. In this case, there's no
-                # need to check that the token is expired because the cache
-                # fetch fails for an expired token. Also, there's no need to
-                # put the token in the cache because it's already in the cache.
+            token_hashes = self._token_hashes(token)
+            cached = self._cache_get_hashes(token_hashes)
 
+            if cached:
                 data = cached
 
                 if self._check_revocations_for_cached:
                     # A token stored in Memcached might have been revoked
                     # regardless of initial mechanism used to validate it,
                     # and needs to be checked.
-                    self._revocations.check(token_ids)
-                self._confirm_token_bind(data, env)
+                    self._revocations.check(token_hashes)
+
+                auth_ref = access.AccessInfo.factory(body=data,
+                                                     auth_token=token)
             else:
                 verified = None
-                # Token wasn't cached. In this case, the token needs to be
-                # checked that it's not expired, and also put in the cache.
+
                 try:
                     if cms.is_pkiz(token):
-                        verified = self._verify_pkiz_token(token, token_ids)
+                        verified = self._verify_pkiz_token(token, token_hashes)
                     elif cms.is_asn1_token(token):
-                        verified = self._verify_signed_token(token, token_ids)
+                        verified = self._verify_signed_token(token,
+                                                             token_hashes)
                 except exceptions.CertificateConfigError:
-                    self._LOG.warn(_LW('Fetch certificate config failed, '
-                                       'fallback to online validation.'))
+                    self._LOG.warning(_LW('Fetch certificate config failed, '
+                                          'fallback to online validation.'))
                 except exc.RevocationListError:
-                    self._LOG.warn(_LW('Fetch revocation list failed, '
-                                       'fallback to online validation.'))
+                    self._LOG.warning(_LW('Fetch revocation list failed, '
+                                          'fallback to online validation.'))
 
                 if verified is not None:
                     data = jsonutils.loads(verified)
-                    expires = _get_token_expiration(data)
-                    _confirm_token_not_expired(expires)
+                    auth_ref = access.AccessInfo.factory(body=data,
+                                                         auth_token=token)
                 else:
-                    data = self._identity_server.verify_token(token)
-                    # No need to confirm token expiration here since
-                    # verify_token fails for expired tokens.
-                    expires = _get_token_expiration(data)
-                self._confirm_token_bind(data, env)
-                self._token_cache.store(token_id, data, expires)
-            return data
+                    auth_ref = self._identity_server.verify_token(token)
+                    if auth_ref:
+                        if auth_ref.version == 'v2.0':
+                            data = {'access': auth_ref}
+                        else:  # it's v3.
+                            data = {'token': auth_ref}
+                    else:
+                        data = None
+
+            # 0 seconds of validity means is it valid right now.
+            if auth_ref.will_expire_soon(stale_duration=0):
+                raise exc.InvalidToken(_('Token authorization failed'))
+
+            if _token_is_v2(data) and not auth_ref.project_id:
+                msg = _('Unable to determine service tenancy.')
+                raise exc.InvalidToken(msg)
+
+            self._confirm_token_bind(data, env)
+
+            if not cached:
+                self._token_cache.store(token_hashes[0], data)
+
+            return auth_ref, data
         except (exceptions.ConnectionRefused, exceptions.RequestTimeout):
             self._LOG.debug('Token validation failure.', exc_info=True)
-            self._LOG.warn(_LW('Authorization failed for token'))
+            self._LOG.warning(_LW('Authorization failed for token'))
             raise exc.InvalidToken(_('Token authorization failed'))
         except exc.ServiceError:
             raise
         except Exception:
             self._LOG.debug('Token validation failure.', exc_info=True)
-            if token_id:
-                self._token_cache.store_invalid(token_id)
-            self._LOG.warn(_LW('Authorization failed for token'))
+            if token_hashes:
+                self._token_cache.store_invalid(token_hashes[0])
+            self._LOG.warning(_LW('Authorization failed for token'))
             raise exc.InvalidToken(_('Token authorization failed'))
 
-    def _build_user_headers(self, auth_ref, token_info):
-        """Convert token object into headers.
-
-        Build headers that represent authenticated user - see main
-        doc info at start of file for details of headers to be defined.
-
-        :param token_info: token object returned by identity
-                           server on authentication
-        :raises exc.InvalidToken: when unable to parse token object
-
-        """
-        roles = ','.join(auth_ref.role_names)
-
-        if _token_is_v2(token_info) and not auth_ref.project_id:
-            raise exc.InvalidToken(_('Unable to determine tenancy.'))
-
-        rval = {
-            'X-Identity-Status': 'Confirmed',
-            'X-Roles': roles,
-        }
-
-        for header_tmplt, attr in six.iteritems(_HEADER_TEMPLATE):
-            rval[header_tmplt % ''] = getattr(auth_ref, attr)
-
-        # Deprecated headers
-        rval['X-Role'] = roles
-        for header_tmplt, attr in six.iteritems(_DEPRECATED_HEADER_TEMPLATE):
-            rval[header_tmplt] = getattr(auth_ref, attr)
-
-        if self._include_service_catalog and auth_ref.has_service_catalog():
-            catalog = auth_ref.service_catalog.get_data()
-            if _token_is_v3(token_info):
-                catalog = _v3_to_v2_catalog(catalog)
-            rval['X-Service-Catalog'] = jsonutils.dumps(catalog)
-
-        return rval
-
-    def _build_service_headers(self, token_info):
-        """Convert token object into service headers.
-
-        Build headers that represent authenticated user - see main
-        doc info at start of file for details of headers to be defined.
-
-        :param token_info: token object returned by identity
-                           server on authentication
-        :raises exc.InvalidToken: when unable to parse token object
-
-        """
-        auth_ref = access.AccessInfo.factory(body=token_info)
-
-        if _token_is_v2(token_info) and not auth_ref.project_id:
-            raise exc.InvalidToken(_('Unable to determine service tenancy.'))
-
-        roles = ','.join(auth_ref.role_names)
-        rval = {
-            'X-Service-Identity-Status': 'Confirmed',
-            'X-Service-Roles': roles,
-        }
-
-        header_type = '-Service'
-        for header_tmplt, attr in six.iteritems(_HEADER_TEMPLATE):
-            rval[header_tmplt % header_type] = getattr(auth_ref, attr)
-
-        return rval
-
-    def _header_to_env_var(self, key):
-        """Convert header to wsgi env variable.
-
-        :param key: http header name (ex. 'X-Auth-Token')
-        :returns: wsgi env variable name (ex. 'HTTP_X_AUTH_TOKEN')
-
-        """
-        return 'HTTP_%s' % key.replace('-', '_').upper()
-
-    def _add_headers(self, env, headers):
-        """Add http headers to environment."""
-        for (k, v) in six.iteritems(headers):
-            env_key = self._header_to_env_var(k)
-            env[env_key] = v
-
-    def _remove_headers(self, env, keys):
-        """Remove http headers from environment."""
-        for k in keys:
-            env_key = self._header_to_env_var(k)
-            try:
-                del env[env_key]
-            except KeyError:
-                pass
-
-    def _get_header(self, env, key, default=None):
-        """Get http header from environment."""
-        env_key = self._header_to_env_var(key)
-        return env.get(env_key, default)
-
     def _invalid_user_token(self, msg=False):
         # NOTE(jamielennox): use False as the default so that None is valid
         if msg is False:
@@ -1090,7 +897,6 @@ class AuthProtocol(object):
 
         cache_kwargs = dict(
             cache_time=int(self._conf_get('token_cache_time')),
-            hash_algorithms=self._conf_get('hash_algorithms'),
             env_cache_name=self._conf_get('cache'),
             memcached_servers=self._conf_get('memcached_servers'),
             use_advanced_pool=self._conf_get('memcache_use_advanced_pool'),
